// 
//  coOccurenceMatrix.cpp
//  IFT630 - Processus concurrents et parallélisme
//  Benoit Gagnon (05 634 691)
//  
//  $Id: coOccurenceMatrix.cpp 2926 2008-04-13 22:45:08Z benoit $
// 

#include "coOccurenceMatrix.h"
#include <iostream>
#include <math.h>

///////////////////////////////////////////////////////////////////////////////

CoOccurenceMatrix::CoOccurenceMatrix(
    const Image< STUInt8 >* img,
    short unsigned int    distance
) : _distance( distance )
{
    assert(img->getColorSpace() == ColorSpaces::Grayscale);

    // Compute de number of neighbour for each angle
    // 0 degrees
    _nbNeighbour[Orient0] = (2*(img->getWidth() - 2*distance) * img->getHeight()
        + 2*distance*img->getHeight());

    // 45 degrees
    _nbNeighbour[Orient45] = (2*(img->getWidth() - 2*distance)
        * (img->getHeight() - 2*distance)
        + 2*distance*(img->getHeight() + img->getWidth())
        - 6*distance*distance);

    // 90 degrees
    _nbNeighbour[Orient90] = (2*(img->getHeight() - 2*distance) * img->getWidth()
        + 2*distance*img->getWidth());

    // 135 degrees
    _nbNeighbour[Orient135] = (2*(img->getWidth() - 2*distance)
        * (img->getHeight() - 2*distance)
        + 2*distance*(img->getHeight() + img->getWidth())
        - 6*distance*distance);

    // Fill tables
    fill( img );
    fillTableMeans();

    // Compute statistics
    computeEnergy();
    computeEntropy();
    computeCorrelation();
    computeInverseDifferenceMoment();
    computeInertia();
    computeClusterShade();
    computeClusterProminence();
}

///////////////////////////////////////////////////////////////////////////////

CoOccurenceMatrix::~CoOccurenceMatrix()
{}

///////////////////////////////////////////////////////////////////////////////

void CoOccurenceMatrix::fill( const Image< STUInt8 >* img )
{
    const int w( img->getWidth() );
    const int h( img->getHeight() );

    for( int x = 0; x < w; ++x ) {
        for( int y = 0; y < h; ++y ) {
            processDirections( img, w, h, x, y );
        }
    }
}

///////////////////////////////////////////////////////////////////////////////

void CoOccurenceMatrix::processDirections(
    const Image< STUInt8>* img,
    const int   w,
    const int   h,
    const int   x,
    const int   y
)
{
    const unsigned short int d( _distance );

    const Image< STUInt8 >::StorageType* src( img->getPixels( x, y ) );
    const Image< STUInt8 >::StorageType ci( *src );

    const unsigned int lineStride( img->getLineStride() );
    
    // 0 degrees
    if( x - d >= 0 )
        ++_table[Orient0][ci][*( src - d )];
    if( x + d < w )
        ++_table[Orient0][ci][*( src + d )];

    // 45 degrees
    if(  x + d < w && y - d >= 0)
        ++_table[Orient45][ci][*( src - ( ( lineStride - 1 ) * d ) )];
    if( x - d >= 0 && y + d < h )
        ++_table[Orient45][ci][*( src + ( ( lineStride + 1 ) * d ) )];

    // 90 degrees
    if( y - d >= 0 )
        ++_table[Orient90][ci][*( src - ( lineStride * d ) )];
    if( y + d < h )
        ++_table[Orient90][ci][*( src + ( lineStride * d ) )];

    // 135 degrees
    if( x - d >= 0 && y - d >= 0 )
        ++_table[Orient135][ci][*( src - ( (lineStride + 1) * d ) )];
    if( x + d < w && y + d < h )
        ++_table[Orient135][ci][*( src + ( (lineStride + 1) * d ) )];
}

///////////////////////////////////////////////////////////////////////////////

void CoOccurenceMatrix::fillTableMeans()
{
    float pMean;
    
    for (int i = 0; i < 256; ++i) {
        for (int j = 0; j < 256; ++j) {
            pMean = 0.0f;
            pMean += _table[  Orient0][i][j] / float( _nbNeighbour[  Orient0] );
            pMean += _table[ Orient45][i][j] / float( _nbNeighbour[ Orient45] );
            pMean += _table[ Orient90][i][j] / float( _nbNeighbour[ Orient90] );
            pMean += _table[Orient135][i][j] / float( _nbNeighbour[Orient135] );
            
            _tableMeans[i][j] = pMean / float( NB_ORIENTATIONS );
        }
    }
}

///////////////////////////////////////////////////////////////////////////////

void CoOccurenceMatrix::computeEnergy()
{
    float enerTemp( 0.0f );
    
    for (int i = 0; i < 256; ++i) {
        for (int j = 0; j < 256; ++j) {
            enerTemp += _tableMeans[i][j] * _tableMeans[i][j];
        }
    }
    
    _energy = enerTemp;
}

///////////////////////////////////////////////////////////////////////////////

void CoOccurenceMatrix::computeEntropy()
{
    float entrTemp( 0.0f );
    
    for (int i = 0; i < 256; ++i) {
        for (int j = 0; j < 256; ++j) {
            if ( _tableMeans[i][j] != 0.0f ) {
                entrTemp += _tableMeans[i][j] * logf( _tableMeans[i][j] );
            }
        }
    }
    
    _entropy = -1.0f * entrTemp;
}

///////////////////////////////////////////////////////////////////////////////

void CoOccurenceMatrix::computeCorrelation()
{
    float u_x( 0.0f );
    float u_y( 0.0f );
    float sigma_x( 0.0f );
    float sigma_y( 0.0f );
    float corrTemp( 0.0f );
    
    // u_x, u_y
    float temp_ux, temp_uy;
    for (int i = 0; i < 256; ++i) {
        temp_ux = 0.0f;
        temp_uy = 0.0f;
        for (int j = 0; j < 256; ++j) {
            temp_ux += _tableMeans[i][j];
            temp_uy += _tableMeans[j][i];
        }
        u_x += i * temp_ux;
        u_y += i * temp_uy;
    }
    
    // sigma_x, sigma_y
    float temp_sx, temp_sy;
    for (int i = 0; i < 256; ++i) {
        temp_sx = 0.0f;
        temp_sy = 0.0f;
        for (int j = 0; j < 256; ++j) {
            temp_sx += _tableMeans[i][j];
            temp_sy += _tableMeans[j][i];
        }
        sigma_x += (i - u_x) * (i - u_x) * temp_sx;
        sigma_y += (i - u_y) * (i - u_y) * temp_sy;
    }

    // Correlation
    for (int i = 0; i < 256; ++i) {
        for (int j = 0; j < 256; ++j) {
            corrTemp += (i - u_x) * (j - u_y) * _tableMeans[i][j];
        }
    }
    
    _correlation = corrTemp / (sigma_x * sigma_y);    
}

///////////////////////////////////////////////////////////////////////////////

void CoOccurenceMatrix::computeInverseDifferenceMoment()
{
    float invDMTemp = 0;
    
    for (int i = 0; i < 256; ++i) {
        for (int j = 0; j < 256; ++j) {
            invDMTemp += _tableMeans[i][j] / (1 + (i-j)*(i-j));
        }
    }
    
    _inverseDifferenceMoment = invDMTemp;
}

///////////////////////////////////////////////////////////////////////////////

void CoOccurenceMatrix::computeInertia()
{
    float inerTemp = 0;
    
    for (int i = 0; i < 256; ++i) {
        for (int j = 0; j < 256; ++j) {
            inerTemp += (i-j)*(i-j) * _tableMeans[i][j];
        }
    }
    
    _inertia = inerTemp;
}

///////////////////////////////////////////////////////////////////////////////

void CoOccurenceMatrix::computeClusterShade()
{
    float u_x( 0.0f );
    float u_y( 0.0f );
    float cShadeTemp( 0.0f );
    
    // u_x, u_y
    float temp_ux, temp_uy;
    for (int i = 0; i < 256; ++i) {
        temp_ux = 0.0f;
        temp_uy = 0.0f;
        for (int j = 0; j < 256; ++j) {
            temp_ux += _tableMeans[i][j];
            temp_uy += _tableMeans[j][i];
        }
        u_x += i * temp_ux;
        u_y += i * temp_uy;
    }

    // Cluster shade
    for (int i = 0; i < 256; ++i) {
        for (int j = 0; j < 256; ++j) {
            float factor = ( i - u_x ) + ( j - u_y );
            cShadeTemp +=  factor * factor * factor * _tableMeans[i][j];
        }
    }

    _clusterShade = cShadeTemp;
}

///////////////////////////////////////////////////////////////////////////////

void CoOccurenceMatrix::computeClusterProminence()
{
    float u_x( 0.0f );
    float u_y( 0.0f );
    float cPromTemp( 0.0f );
    
    // u_x, u_y
    float temp_ux, temp_uy;
    for (int i = 0; i < 256; ++i) {
        temp_ux = 0.0f;
        temp_uy = 0.0f;
        for (int j = 0; j < 256; ++j) {
            temp_ux += _tableMeans[i][j];
            temp_uy += _tableMeans[j][i];
        }
        u_x += i * temp_ux;
        u_y += i * temp_uy;
    }
    
    // Cluster prominence
    for (int i = 0; i < 256; ++i) {
        for (int j = 0; j < 256; ++j) {
            float factor = ( i - u_x ) + ( j - u_y );
            cPromTemp +=  factor * factor * factor * factor * _tableMeans[i][j];
        }
    }
    
    _clusterProminence = cPromTemp;
}
