// 
//  collectionReport.cpp
//  IFT630 - Processus concurrents et parallélisme
//  Benoit Gagnon (05 634 691)
//  
//  $Id: collectionReport.cpp 2926 2008-04-13 22:45:08Z benoit $
// 

#include "collectionReport.h"
#include "osUtils.h"
#include <iostream>
#include <fstream>
#include <algorithm>
#include <functional>

///////////////////////////////////////////////////////////////////////////////

struct DescriptorWriter
{
    DescriptorWriter( std::ostream& os ) :_os( os ){}
    void operator()( const ImageFeature::Descriptor& obj ){ obj.write( _os ); }
    std::ostream& _os;
};

///////////////////////////////////////////////////////////////////////////////

struct ImageReportWriter
{
    ImageReportWriter( std::ostream& os ) :_os( os ){}
    void operator()( const CollectionReport::ReportPtr& report ){
        report->write( _os );
    }
    std::ostream& _os;
};

///////////////////////////////////////////////////////////////////////////////

struct RankerWrapper
{
    RankerWrapper( const ImageRanker& r ) : _r( r ){}
    SearchResult operator()(
        const shared_ptr< const ImageReport >& obj
    ) const
    { return SearchResult( obj, _r.rank( obj ) ); }
    const ImageRanker& _r;
};

///////////////////////////////////////////////////////////////////////////////

struct PostProcessorWrapper
{
    PostProcessorWrapper( ImageReportPostProcessor& p )
    : _p( p )
    {}
    void operator()( const shared_ptr< ImageReport >& obj )
    { _p.process( *obj ); }
    ImageReportPostProcessor& _p;
};

///////////////////////////////////////////////////////////////////////////////

CollectionReport::Header::Header()
{}

///////////////////////////////////////////////////////////////////////////////

CollectionReport::Header::Header( std::istream& is )
{
    // Magic number ( 2 bytes )
    assert(is.good());
    char magic[2];
    is.read( magic, 2 );
    if( magic[0] != 'S' || magic[1] != 'B' ) {
        throw "Incorrect magic number";
    }

    // Endianness ( 1 byte )
    char endianness = is.get();
    if( endianness != OSUtils::getEndianness() ) {
        throw "File endianness does not match platform's";
    }

    // Nb of features ( 1 byte )
    char nbFeatures = is.get();
    assert( nbFeatures > 0 );
    _descriptors.reserve( nbFeatures );

    // Read all feature descriptors ( 16 bytes each )
    for( int i = 0; i < nbFeatures; ++i ) {
        _descriptors.push_back( ImageFeature::Descriptor( is ) );
    }
}

///////////////////////////////////////////////////////////////////////////////

void CollectionReport::Header::add( const ImageFeature::Descriptor& desc )
{
    _descriptors.push_back( desc );
}

///////////////////////////////////////////////////////////////////////////////

void CollectionReport::Header::write( std::ostream& os ) const
{
    // Magic number ( 2 bytes )
    os << 'S' << 'B';

    // Write endianness ( 1 byte )
    os << static_cast< unsigned char>( OSUtils::getEndianness() );

    // Write nb of features ( 1 byte )
    os << static_cast< unsigned char >( _descriptors.size() );

    // Write all feature descriptors ( 16 bytes each )
    std::for_each(
        _descriptors.begin(),
        _descriptors.end(),
        DescriptorWriter( os )
    );
}

///////////////////////////////////////////////////////////////////////////////

CollectionReport::Header::FeaturePair
CollectionReport::Header::getFeature(
    ImageFeature::Type      type,
    ColorSpaces::ColorSpace colorSpace /* = ColorSpaces::UNKNOWN */
) const
{
    std::vector< ImageFeature::Descriptor >::const_iterator
    it( _descriptors.begin() );
    while( it != _descriptors.end() ) {
        if( it->_type == type ) {
            // Check the colorspace if necessary
            if( colorSpace == ColorSpaces::UNKNOWN ||
                it->_colorSpace == colorSpace ) {
                return FeaturePair( it, true );
            }
        }
        ++it;
    }
    return FeaturePair( it, false );
}

///////////////////////////////////////////////////////////////////////////////

CollectionReport::CollectionReport()
{}

///////////////////////////////////////////////////////////////////////////////

CollectionReport::~CollectionReport()
{}

///////////////////////////////////////////////////////////////////////////////

bool CollectionReport::insert( ReportPtr imgReport )
{
    //TODO Lock a mutex if the app becomes multithreaded
    return _reports.insert( imgReport ).second;
}

///////////////////////////////////////////////////////////////////////////////

void CollectionReport::write( std::ostream& os ) const
{
    // Write image reports
    std::for_each(
        _reports.begin(),
        _reports.end(),
        ImageReportWriter( os )
    );
}

///////////////////////////////////////////////////////////////////////////////

unsigned int CollectionReport::getNbImages() const
{
    return static_cast< unsigned int >( _reports.size() );
}

///////////////////////////////////////////////////////////////////////////////

void CollectionReport::collect(
    std::multiset< SearchResult >& results,
    const ImageRanker&             ranker
) const
{
    std::transform(
       _reports.begin(),
       _reports.end(),
       std::inserter( results, results.begin() ),
       RankerWrapper( ranker )
    );
}

///////////////////////////////////////////////////////////////////////////////

void CollectionReport::postProcess(
    ImageReportPostProcessor& postProcessor
)
{
    //TODO Lock a mutex if the app becomes multithreaded
    std::for_each(
        _reports.begin(),
        _reports.end(),
        PostProcessorWrapper( postProcessor )
    );
}
