// 
//  colorUtil.cpp
//  IFT630 - Processus concurrents et parallélisme
//  Benoit Gagnon (05 634 691)
//  
//  $Id: colorUtil.cpp 2926 2008-04-13 22:45:08Z benoit $
// 

#include "colorUtil.h"
#include "colorConverter.h"
#include "image.h"

#include "convertRGBtoCMY.h"
#include "convertRGBtoGrayscale.h"
#include "convertRGBtoXYZ.h"
#include "convertXYZtoLab.h"
#include "convertLabtoHVC.h"
#include <assert.h>

///////////////////////////////////////////////////////////////////////////////

ColorConverter*
ColorUtil::_converters[NB_COLOR_SPACES][NB_PIXEL_DEPTHS]
                      [NB_COLOR_SPACES][NB_PIXEL_DEPTHS];
bool ColorUtil::_initialized( false );

///////////////////////////////////////////////////////////////////////////////

void ColorUtil::initialize()
{
    assert( !_initialized );

    // RGB -> CMY
    _converters[RGB][STUInt8][CMY][STUInt8] =
        new ConvertRGBtoCMY< STUInt8, STUInt8 >();
    _converters[RGB][STUInt16][CMY][STUInt16] =
        new ConvertRGBtoCMY< STUInt16, STUInt16 >();

    // RGB -> Grayscale
    _converters[RGB][STUInt8][Grayscale][STUInt8] =
        new ConvertRGBtoGrayscale< STUInt8, STUInt8 >();
    _converters[RGB][STUInt16][Grayscale][STUInt16] =
        new ConvertRGBtoGrayscale< STUInt16, STUInt16 >();

    // RGB -> XYZ
    _converters[RGB][STUInt8][XYZ][STFloat32] =
        new ConvertRGBtoXYZ< STUInt8, STFloat32 >();
    _converters[RGB][STUInt16][XYZ][STFloat32] =
        new ConvertRGBtoXYZ< STUInt16, STFloat32 >();
    _converters[RGB][STFloat32][XYZ][STFloat32] =
        new ConvertRGBtoXYZ< STFloat32, STFloat32 >();
    
    // XYZ -> Lab
    _converters[XYZ][STFloat32][Lab][STFloat32] =
        new ConvertXYZtoLab< STFloat32, STFloat32 >();
    
    // Lab -> HVC
    _converters[Lab][STFloat32][HVC][STFloat32] =
        new ConvertLabtoHVC< STFloat32, STFloat32 >();

    // Adapters
    
    CompositeColorConverter* adapter = 0;

    // RGB (STUInt8) -> HVC (STFloat32)
    adapter = new CompositeColorConverter();
    adapter->linkConverter(
        getDirectConverter( RGB,   STUInt8, XYZ, STFloat32 ) );
    adapter->linkConverter(
        getDirectConverter( XYZ, STFloat32, Lab, STFloat32 ) );
    adapter->linkConverter(
        getDirectConverter( Lab, STFloat32, HVC, STFloat32 ) );
    _converters[RGB][STUInt8][HVC][STFloat32] = adapter;

    // RGB (STUInt16) -> HVC (STFloat32)
    adapter = new CompositeColorConverter();
    adapter->linkConverter( 
        getDirectConverter( RGB,  STUInt16, XYZ, STFloat32 ) );
    adapter->linkConverter( 
        getDirectConverter( XYZ, STFloat32, Lab, STFloat32 ) );
    adapter->linkConverter( 
        getDirectConverter( Lab, STFloat32, HVC, STFloat32 ) );
    _converters[RGB][STUInt16][HVC][STFloat32] = adapter;
    
    // RGB (STFloat32) -> HVC (STFloat32)
    adapter = new CompositeColorConverter();
    adapter->linkConverter( 
        getDirectConverter( RGB,  STFloat32, XYZ, STFloat32 ) );
    adapter->linkConverter( 
        getDirectConverter( XYZ, STFloat32, Lab, STFloat32 ) );
    adapter->linkConverter( 
        getDirectConverter( Lab, STFloat32, HVC, STFloat32 ) );
    _converters[RGB][STFloat32][HVC][STFloat32] = adapter;
    
    _initialized = true;
}

///////////////////////////////////////////////////////////////////////////////

void ColorUtil::finalize()
{
    assert( _initialized );

    // RGB -> CMY
    delete _converters[RGB][STUInt8][CMY][STUInt8];
    delete _converters[RGB][STUInt16][CMY][STUInt16];

    // RGB -> Grayscale
    delete _converters[RGB][STUInt8][Grayscale][STUInt8];
    delete _converters[RGB][STUInt16][Grayscale][STUInt16];

    // RGB -> XYZ
    delete _converters[RGB][STUInt8][XYZ][STFloat32];
    delete _converters[RGB][STUInt16][XYZ][STFloat32];

    // XYZ -> Lab
    delete _converters[XYZ][STFloat32][Lab][STFloat32];
    
    // Lab -> HVC
    delete _converters[Lab][STFloat32][HVC][STFloat32];
    
    // Adapters
    
    // RGB -> HVC
    delete _converters[RGB][STUInt8][HVC][STFloat32];
    delete _converters[RGB][STUInt16][HVC][STFloat32];
    delete _converters[RGB][STFloat32][HVC][STFloat32];

    _initialized = false;
}

///////////////////////////////////////////////////////////////////////////////

ImageBase* ColorUtil::convert(
    const ImageBase* srcImg,
    ColorSpace       targetSpace,
    PixelDepth       targetDepth
)
{
    assert( _initialized );

    ColorConverter* converter = 0;
    converter = _converters[srcImg->getColorSpace()][srcImg->getPixelDepth()]
                           [targetSpace][targetDepth];
    if( !converter ) {
        throw "No converter found. ";
    }

    return converter->convert( srcImg );
}

///////////////////////////////////////////////////////////////////////////////

DirectColorConverter* ColorUtil::getDirectConverter(
    ColorSpace srcSpace,
    PixelDepth srcDepth,
    ColorSpace targetSpace,
    PixelDepth targetDepth
)
{
    DirectColorConverter* converter = 0;
    converter = dynamic_cast< DirectColorConverter* >(
                    _converters[   srcSpace][   srcDepth]
                               [targetSpace][targetDepth] );

    return converter;
}

