// 
//  correlogram.cpp
//  IFT630 - Processus concurrents et parallélisme
//  Benoit Gagnon (05 634 691)
//  
//  $Id: correlogram.cpp 2926 2008-04-13 22:45:08Z benoit $
// 

#include "indexedImage.h"
#include "pixelIndexer.h"
#include "correlogram.h"
#include "histogram.h"
#include <cstdlib>
#include <memory.h>
#include <math.h>

#define MIN( a, b ) ( a < b  ? a : b )
#define MAX( a, b ) ( a > b  ? a : b )

///////////////////////////////////////////////////////////////////////////////

Correlogram::Correlogram(
    const IndexedImage*     img,
    unsigned short          distance
) : _distance( distance ),
    _indexer( img->getPixelIndexer() )
{
    _nbColors = _indexer->getNbBins();

    // Allocate memory for occurence matrix
    _occurences = new StorageType*[_nbColors];
    for( unsigned int i = 0; i < _nbColors; ++i ) {
        _occurences[i] = new StorageType[_nbColors];
    }

    // Allocate memory for travel paths vector
    _paths = new StorageType[_nbColors];
    
    for ( unsigned int i = 0; i < _nbColors; ++i ) {
        for ( unsigned int j = 0; j < _nbColors; ++j ) {
            _occurences[i][j] = 0;
        }
        _paths[i] = 0;
    }

    // Fill matrix and vector
    fill( img );

    // Compute statistics
    computeContrast();
    computeUniformity();
    computeInverseDifferenceMoment();
    computeEntropy();
}

///////////////////////////////////////////////////////////////////////////////

Correlogram::~Correlogram()
{
    // Free memory used by matrices

    for( unsigned long i = 0; i < _nbColors; i++ ) {
        delete[] _occurences[i];
    }
    delete[] _occurences;
    delete[] _paths;
}

///////////////////////////////////////////////////////////////////////////////

unsigned long Correlogram::getNbBins() const {
    return _nbColors;
}

///////////////////////////////////////////////////////////////////////////////

unsigned short Correlogram::getDistance() const {
    return _distance;
}

///////////////////////////////////////////////////////////////////////////////

Correlogram::StorageType Correlogram::getOccurence(
    unsigned long ci,
    unsigned long cj
) const
{
    return _occurences[ci][cj];
}

///////////////////////////////////////////////////////////////////////////////

float Correlogram::getRatio(
    unsigned long ci,
    unsigned long cj
) const
{
    if (_paths[ci] == 0) {
        return 0.0f;
    }
    else {
        float temp_f = (float)(_occurences[ci][cj]) / (float)(_paths[ci]);
        int temp_i = (int)(1000*temp_f);
        return (float)(temp_i/1000.0f);
    }
}

///////////////////////////////////////////////////////////////////////////////

void Correlogram::fill( const IndexedImage* img )
{
    const unsigned int w( img->getWidth() );
    const unsigned int h( img->getHeight() );
    const IndexedImage::StorageType* src( img->getPixels() );
    for( unsigned int x = 0; x < w; ++x ) {
        for( unsigned int y = 0; y < h; ++y ) {
             processVicinity( img, w, h, x, y, *(src++) );
        }
    }
}

///////////////////////////////////////////////////////////////////////////////

void Correlogram::processVicinity(
    const IndexedImage*              img,
    const int                        w,
    const int                        h,
    const long                       i,
    const long                       j,
    IndexedImage::StorageType        ci
)
{
    // (1)---->(2)
    //  ^       |
    //  |   o   |
    //  |       V
    // (4)<----(3)

    const short d( _distance );

    long x, y;

    const IndexedImage::StorageType* src;
    const unsigned int lineStride( img->getLineStride() );
    
    bool topBorder = (MIN( j - d, 0 ) == 0);
    bool rightBorder = (MAX( i + d, ( w - 1 ) ) == ( w - 1 ));
    bool bottomBorder = (MAX( j + d, ( h - 1 ) ) == ( h - 1 ));
    bool leftBorder = (MIN( i - d, 0 ) == 0);

    long xMin, xMax, yMin, yMax;
    xMin = MAX( i - d, 0 );
    xMax = MIN( i + d, w - 1 );
    yMin = MAX( j - d, 0 );
    yMax = MIN( j + d, h - 1 );
    
    long xMin_(xMin), xMax_(xMax);
    if ( !rightBorder ) {
        xMax_ = xMax + 1;
    }
    if ( !leftBorder ) {
        xMin_ = xMin - 1;
    }

    StorageType* occPtr( _occurences[ci] );
    StorageType  path = 0;

    // (1) top-left corner + top border
    if( topBorder ) {
        src = img->getPixels( xMin, yMin );
        for( x = xMin_; x < xMax_; ++x ) {
            ++occPtr[*src];
            ++path;
            ++src;
        }
    }

    // (2) top-right corner + right border
    if( rightBorder ) {
        src = img->getPixels( xMax, yMin );
        for( y = yMin; y < yMax; ++y ) {
            ++occPtr[*src];
            ++path;
            src += lineStride;
        }
    }

    // (3) bottom-right corner + bottom border
    if( bottomBorder ) {
        src = img->getPixels( xMax, yMax );
        for( x = xMax_; x > xMin_; --x ) {
            ++occPtr[*src];
            ++path;
            --src;
        }
    }

    // (4) bottom-right corner + left border
    if( leftBorder ) {
        src = img->getPixels( xMin, yMax );
        for( y = yMax; y > yMin; --y ) {
            ++occPtr[*src];
            ++path;
            src -= lineStride;
        }
    }

    _paths[ci] += path;
}

///////////////////////////////////////////////////////////////////////////////

shared_ptr< const PixelIndexer > Correlogram::getPixelIndexer() const
{
    return _indexer;
}

///////////////////////////////////////////////////////////////////////////////

void Correlogram::computeContrast()
{
    float contTemp = 0.0f;
    
    for ( unsigned int i = 0; i < getNbBins(); ++i ) {
        for ( unsigned int j = 0; j < getNbBins(); ++j ) {
            contTemp += _indexer->getDist(i, j) * getRatio(i, j);
        }
    }
    
    _contrast = contTemp;
}

///////////////////////////////////////////////////////////////////////////////

void Correlogram::computeUniformity()
{
    float unifTemp = 0.0f;
    
    for ( unsigned int i = 0; i < getNbBins(); ++i ) {
        for ( unsigned int j = 0; j < getNbBins(); ++j) {
            unifTemp += getRatio( i, j ) * getRatio( i, j );
        }
    }
    
    _uniformity = unifTemp;
}

///////////////////////////////////////////////////////////////////////////////

void Correlogram::computeInverseDifferenceMoment()
{
    float invDMTemp = 0.0f;
    
    for ( unsigned int i = 0; i < getNbBins(); ++i ) {
        for ( unsigned int j = 0; j < getNbBins(); ++j) {
            invDMTemp += getRatio( i, j ) 
                / (1 + _indexer->getDist( i, j ) * _indexer->getDist( i, j ) );
        }
    }
    
    _inverseDifferenceMoment = invDMTemp;
}

///////////////////////////////////////////////////////////////////////////////

void Correlogram::computeEntropy()
{
    float entrTemp = 0.0f;
    
    for ( unsigned int i = 0; i < getNbBins(); ++i ) {
        for ( unsigned int j = 0; j < getNbBins(); ++j ) {
            if ( getRatio( i, j ) != 0.0f ) {
                entrTemp += getRatio( i, j ) * logf( getRatio( i, j ) );
            }
        }
    }
    
    _entropy = entrTemp;
}
