// 
//  digestQueue.cpp
//  IFT630 - Processus concurrents et parallélisme
//  Benoit Gagnon (05 634 691)
//  
//  $Id: digestQueue.cpp 2940 2008-04-15 03:48:29Z benoit $
// 

#include "digestQueue.h"
#include "digestThread.h"
#include "collectionReport.h"
#include "textureFeatureRangeFinder.h"
#include "textureFeatureNormalizer.h"
#include "imageReportFactory.h"
#include "osUtils.h"

///////////////////////////////////////////////////////////////////////////////

DigestQueue::DigestQueue(
    const FSUtils::FileList&                            fileList,
    const shared_ptr< const CollectionReport::Header >& header
)
: _fileList( fileList ),
  _collection( new CollectionReport() ),
  _header( header )
{
    // Number of threads to spawn
    const int nbThreads = OSUtils::getCPUcount();
    
    // Instantiate factory
    shared_ptr< ImageReportFactory > imgReportFactory(
        new ImageReportFactory( _header ) );
    
    // Create N threads
    for( int i = 0; i < nbThreads; ++i )
        _threads.push_back( shared_ptr< DigestThread >(
            new DigestThread( imgReportFactory, this ) ) );
    
    // Start all threads
    std::vector< shared_ptr< DigestThread > >::const_iterator it;
    it = _threads.begin();
    while( it != _threads.end() ) {
        if( (*it)->Create() != wxTHREAD_NO_ERROR )
            wxLogError( _T( "Can't create thread!" ) );
        (*it)->Run();
        ++it;
    }
}

///////////////////////////////////////////////////////////////////////////////

bool DigestQueue::getNextFile( std::wstring& file )
{
    wxMutexLocker lock( _fileListmutex );
    
    if( _fileList.empty() ) {
        return false;
    }
    else {
        file = *_fileList.begin();
        _fileList.pop_front();
        return true;
    }
}

///////////////////////////////////////////////////////////////////////////////

size_t DigestQueue::getSize() const
{
    wxMutexLocker lock( _fileListmutex );
    return _fileList.size();
}

///////////////////////////////////////////////////////////////////////////////

bool DigestQueue::addReport( const shared_ptr< ImageReport >& imgReport )
{
    wxMutexLocker lock( _collectionMutex );
    return _collection->insert( imgReport );
}

///////////////////////////////////////////////////////////////////////////////

shared_ptr< CollectionReport > DigestQueue::getCollectionReport()
{
    joinAll();
    
    // Post-processing
    if ( _header->getFeature( ImageFeature::CORRELOGRAM).second ) {
        TextureFeatureRangeFinder textureFRF(ImageFeature::CORRELOGRAM);
        _collection->postProcess(textureFRF);
        TextureFeatureNormalizer textureFN(
            ImageFeature::CORRELOGRAM,
            textureFRF.getMinMaxPairs() );
        _collection->postProcess(textureFN);
    }
    
    if ( _header->getFeature( ImageFeature::CO_OCCURENCE).second ) {
        TextureFeatureRangeFinder textureFRF(ImageFeature::CO_OCCURENCE);
        _collection->postProcess(textureFRF);
        TextureFeatureNormalizer textureFN(
            ImageFeature::CO_OCCURENCE,
            textureFRF.getMinMaxPairs() );
        _collection->postProcess(textureFN);
    }
    
    shared_ptr< CollectionReport > report( _collection );
    _collection.reset( new CollectionReport() );
    return report;
}

///////////////////////////////////////////////////////////////////////////////

void DigestQueue::abort()
{
    std::vector< shared_ptr< DigestThread > >::const_iterator it;
    it = _threads.begin();
    while( it != _threads.end() ) {
        (*it)->abort();
        ++it;
    }
    
    joinAll();
}

///////////////////////////////////////////////////////////////////////////////

void DigestQueue::joinAll()
{
    // Join all threads
    std::vector< shared_ptr< DigestThread > >::const_iterator it;
    it = _threads.begin();
    while( it != _threads.end() ) {
        (*it)->Wait();
        ++it;
    }
    
    _threads.clear();
}
