// 
//  histogramFeature.cpp
//  IFT630 - Processus concurrents et parallélisme
//  Benoit Gagnon (05 634 691)
//  
//  
// 

#include "histogramFeature.h"
#include "histogram.h"
#include "pixelIndexer.h"
#include <ostream>
#include <algorithm>
#include <numeric>
#include <iterator>
#include <math.h>

///////////////////////////////////////////////////////////////////////////////

template< typename T >
struct Mininum
{
    Mininum(){}
    T operator()( const T& a, const T& b ){ return a < b ? a : b;}
};

///////////////////////////////////////////////////////////////////////////////

template< typename T >
struct SquareDifference
{
    SquareDifference(){}
    T operator()( const T& a, const T& b ){ const T r( a - b ); return r*r; }
};

///////////////////////////////////////////////////////////////////////////////

template< typename T >
struct AbsDifference
{
    AbsDifference(){}
    T operator()( const T& a, const T& b ){ return fabs( a - b ); }
};

///////////////////////////////////////////////////////////////////////////////

ImageFeature::Type HistogramFeature::getType() const
{
    return HISTOGRAM;
}

///////////////////////////////////////////////////////////////////////////////

ColorSpaces::ColorSpace HistogramFeature::getColorSpace() const
{
    return _indexer->getTargetColorSpace();
}

///////////////////////////////////////////////////////////////////////////////

HistogramFeature::HistogramFeature( const shared_ptr< const Histogram >& histo )
: _bins(),
  _indexer( histo->getPixelIndexer() )
{
    // Pre-allocate memory
    _bins.reserve( histo->getNbBins() );

    // Normalize every entry
    // and store them in the float array
    std::transform(
        histo->begin(), histo->end(),
        std::back_inserter( _bins ),
        std::bind2nd( std::divides< StorageType >(), histo->getTotalCount() )
    );
}

///////////////////////////////////////////////////////////////////////////////

HistogramFeature::HistogramFeature(
    std::istream&                    is,
    shared_ptr< const PixelIndexer > indexer
) : _bins( indexer->getNbBins(), 0.0f ),
    _indexer( indexer )
{
    // Read all the histogram values in a block
    is.read( reinterpret_cast< char* >( &( _bins[0] ) ),
             sizeof( StorageType ) * indexer->getNbBins() );
}

///////////////////////////////////////////////////////////////////////////////

HistogramFeature::~HistogramFeature()
{}

///////////////////////////////////////////////////////////////////////////////

void HistogramFeature::write( std::ostream& os ) const
{
    // Dump the vector to the file
    os.write(
        reinterpret_cast< const char* >( &_bins[0] ),
        _bins.size() * sizeof( StorageType )
    );
}

///////////////////////////////////////////////////////////////////////////////

float HistogramFeature::euclideanDistance( const HistogramFeature& left ) const
{
    assert( left._bins.size() == _bins.size() );
    assert( left._indexer == _indexer );

    std::vector< StorageType > sqrDiffs;
    sqrDiffs.reserve( _bins.size() );

    // Collect the square value of the difference between each bin
    std::transform(
        left._bins.begin(),
        left._bins.end(),
        _bins.begin(),
        std::back_inserter( sqrDiffs ),
        SquareDifference< StorageType >()
    );

    // Square root of the sum of the square differences
    return sqrtf( std::accumulate( sqrDiffs.begin(), sqrDiffs.end(), 0.0f ) );
}

///////////////////////////////////////////////////////////////////////////////

float HistogramFeature::weightedDistance( const HistogramFeature& left ) const
{
    assert( left._bins.size() == _bins.size() );
    assert( left._indexer == _indexer );

    std::vector< StorageType >  dists( _bins.size(), 0.0f );
    
    for( unsigned int ci = 0; ci < _bins.size(); ++ci ) {
        for( unsigned int cj = 0; cj < _bins.size(); ++cj ) {
            float diffi = fabs(_bins[ci] - left._bins[ci]);
            float diffj = fabs(_bins[cj] - left._bins[cj]);
            dists[ci] +=
                ( 1.0f - _indexer->getDist( ci, cj ) ) * diffi * diffj;
        }
    }
    
    // Square root of the sum of the square differences
    return sqrtf( std::accumulate( dists.begin(), dists.end(), 0.0f ) );
}

///////////////////////////////////////////////////////////////////////////////

float HistogramFeature::intersectionDistance(
    const HistogramFeature& left
) const
{
    assert( left._bins.size() == _bins.size() );
    assert( left._indexer == _indexer );

    std::vector< StorageType > minimas;
    minimas.reserve( _bins.size() );

    // Collect the min. value for every bin
    std::transform(
        left._bins.begin(),
        left._bins.end(),
        _bins.begin(),
        std::back_inserter( minimas ),
        Mininum< StorageType >()
    );

    // Sum the minimum values
    // note: Multiply by -1 because the larger the value, the better the score
    return -1.0f * std::accumulate( minimas.begin(), minimas.end(), 0.0f );
}

///////////////////////////////////////////////////////////////////////////////

float HistogramFeature::manhattanDistance( const HistogramFeature& left ) const
{
    assert( left._bins.size() == _bins.size() );
    assert( left._indexer == _indexer );

    std::vector< StorageType > absDiffs;
    absDiffs.reserve( _bins.size() );

    // Collect the absolute value of the difference between each bin
    std::transform(
        left._bins.begin(),
        left._bins.end(),
        _bins.begin(),
        std::back_inserter( absDiffs ),
        AbsDifference< StorageType >()
    );

    // Sum the absolute differences
    return std::accumulate( absDiffs.begin(), absDiffs.end(), 0.0f );
}
