// 
//  imageReportFactory.cpp
//  IFT630 - Processus concurrents et parallélisme
//  Benoit Gagnon (05 634 691)
//  
//  $Id: imageReportFactory.cpp 2928 2008-04-13 22:48:01Z benoit $
// 

#include "imageReportFactory.h"
#include "staticPixelIndexer.h"
#include "kMeansPixelIndexer.h"
#include "ppmImageFile.h"
#include "histogramFeature.h"
#include "histogram.h"
#include "correlogramFeature.h"
#include "coOccurenceMatrixFeature.h"
#include "indexedImage.h"
#include "image.h"
#include "stringUtils.h"
#include "scoped_ptr.h"
#include <iostream>

///////////////////////////////////////////////////////////////////////////////

ImageReportFactory::ImageReportFactory(
    shared_ptr< const CollectionReport::Header > header
)
: _header( header ),
  _nbFeatures( 0 ),
  _corrDistances( 0 ),
  _coOccDistances( 0 )

{
    CollectionReport::Header::FeaturePair fp;

    // RGB histo
    fp = header->getFeature( ImageFeature::HISTOGRAM, ColorSpaces::RGB );
    if( fp.second ) {
        _rgbHistogramIndexer = shared_ptr< PixelIndexer >(
            StaticPixelIndexer::fromParams( fp.first->_parameters ) );
        ++_nbFeatures;
    }

    // HVC histo
    fp = header->getFeature( ImageFeature::HISTOGRAM, ColorSpaces::HVC );
    if( fp.second ) {
        _hvcHistogramIndexer = shared_ptr< PixelIndexer >(
            KMeansPixelIndexer::fromParams( fp.first->_parameters ) );
        ++_nbFeatures;
    }

    // Correlogram(s)
    fp = header->getFeature( ImageFeature::CORRELOGRAM );
    if( fp.second ) {
        _corrDistances = reinterpret_cast< const unsigned char* >(
            &( fp.first->_parameters ) )[7];
        ++_nbFeatures;
        switch( fp.first->_colorSpace ) {
            case ColorSpaces::RGB: {
                _correlogramIndexer = shared_ptr< PixelIndexer >(
                    StaticPixelIndexer::fromParams( fp.first->_parameters ) );
            } break;
            case ColorSpaces::HVC: {
                _correlogramIndexer = shared_ptr< PixelIndexer >(
                    KMeansPixelIndexer::fromParams( fp.first->_parameters ) );
            } break;
        }
    }

    // CoOccurence matrices
    fp = header->getFeature( ImageFeature::CO_OCCURENCE );
    if( fp.second ) {
        _coOccDistances = reinterpret_cast< const unsigned char* >(
            &( fp.first->_parameters ) )[7];
        ++_nbFeatures;
    }

    assert( _nbFeatures > 0 );
}

///////////////////////////////////////////////////////////////////////////////

ImageReportFactory::ImageReportFactory(
    shared_ptr< const CollectionReport::Header > header,
    std::istream&                                is
)
: _header( header ),
  _nbFeatures( 0 ),
  _corrDistances( 0 ),
  _coOccDistances( 0 )
  
{
    CollectionReport::Header::FeaturePair fp;

    // RGB histo
    fp = header->getFeature( ImageFeature::HISTOGRAM, ColorSpaces::RGB );
    if( fp.second ) {
        _rgbHistogramIndexer = shared_ptr< PixelIndexer >(
            StaticPixelIndexer::fromParams( fp.first->_parameters ) );
        ++_nbFeatures;
    }

    // HVC histo
    fp = header->getFeature( ImageFeature::HISTOGRAM, ColorSpaces::HVC );
    if( fp.second ) {
        _hvcHistogramIndexer = shared_ptr< PixelIndexer >(
            KMeansPixelIndexer::fromParams( is, fp.first->_parameters ) );
        ++_nbFeatures;
    }

    // Correlogram(s)
    fp = header->getFeature( ImageFeature::CORRELOGRAM );
    if( fp.second ) {
        _corrDistances = reinterpret_cast< const unsigned char* >(
            &( fp.first->_parameters ) )[7];
        ++_nbFeatures;
        switch( fp.first->_colorSpace ) {
            case ColorSpaces::RGB: {
                _correlogramIndexer = shared_ptr< PixelIndexer >(
                    StaticPixelIndexer::fromParams( fp.first->_parameters ) );
            } break;
            case ColorSpaces::HVC: {
                _correlogramIndexer = shared_ptr< PixelIndexer >(
                  KMeansPixelIndexer::fromParams( is, fp.first->_parameters ) );
            } break;
        }
    }

    // CoOccurence matrices
    fp = header->getFeature( ImageFeature::CO_OCCURENCE );
    if( fp.second ) {
        _coOccDistances = reinterpret_cast< const unsigned char* >(
            &( fp.first->_parameters ) )[7];
        ++_nbFeatures;
    }

    assert( _nbFeatures > 0 );
}

///////////////////////////////////////////////////////////////////////////////

ImageReportFactory::~ImageReportFactory()
{}

///////////////////////////////////////////////////////////////////////////////

shared_ptr< ImageReport >
ImageReportFactory::create( const std::wstring& filename ) const
{
    shared_ptr< ImageReport > imgReport(
        new ImageReport( filename, _nbFeatures ) );

    scoped_ptr< ImageFile > imageFile( ImageFile::create( filename ) );

    // Read the image, let potential exception bubble up
    scoped_ptr< Image< STUInt8 > > img( dynamic_cast< Image< STUInt8 >* >(
        imageFile->read() ) );

    //FIXME See if we can re-use the indexed img from the correlogram
    // and speed things up
    if( _rgbHistogramIndexer ) {
        imgReport->add( shared_ptr< HistogramFeature >(
            new HistogramFeature( shared_ptr< Histogram >(
                new Histogram( img.get(), _rgbHistogramIndexer ) ) ) ) );
    }

    //FIXME See if we can re-use the indexed img from the correlogram
    // and speed things up
    if( _hvcHistogramIndexer ) {
        imgReport->add( shared_ptr< HistogramFeature >(
            new HistogramFeature( shared_ptr< Histogram >(
                new Histogram( img.get(), _hvcHistogramIndexer ) ) ) ) );
    }

    if( _coOccDistances ) {
        scoped_ptr< Image< STUInt8 > > imgGray(
            dynamic_cast< Image< STUInt8 >* >( ColorUtil::convert(
                img.get(),
                ColorSpaces::Grayscale,
                STUInt8 ) ) );
        imgReport->add( shared_ptr< CoOccurenceMatrixFeature> (
            new CoOccurenceMatrixFeature( imgGray.get(), _coOccDistances ) )
        );
    }

    if( _corrDistances ) {
        scoped_ptr< IndexedImage > indexedImg(
            new IndexedImage( img.get(), _correlogramIndexer ) );
        imgReport->add( shared_ptr< CorrelogramFeature >(
            new CorrelogramFeature( indexedImg.get(), _corrDistances ) )
        );
    }

    return imgReport;
}

///////////////////////////////////////////////////////////////////////////////

shared_ptr< ImageReport >
ImageReportFactory::create( std::istream& is ) const
{
    // Read the filename up to \0
    // Maximum of 1024 chars
    char filename[1024];
    is.getline( filename, 1024, '\0' );
    
    shared_ptr< ImageReport > imgReport(
        new ImageReport( str2wstr( filename ), _nbFeatures ) );

    if( _rgbHistogramIndexer ) {
        imgReport->add( shared_ptr< HistogramFeature >(
            new HistogramFeature( is, _rgbHistogramIndexer ) ) );
    }

    if( _hvcHistogramIndexer ) {
        imgReport->add( shared_ptr< HistogramFeature >(
            new HistogramFeature( is, _hvcHistogramIndexer ) ) );
    }

    if( _coOccDistances ) {
        imgReport->add( shared_ptr< CoOccurenceMatrixFeature>(
            new CoOccurenceMatrixFeature( is ) ) );
    }

    if( _corrDistances ) {
        imgReport->add( shared_ptr< CorrelogramFeature >(
            new CorrelogramFeature( is,
                _correlogramIndexer->getTargetColorSpace() ) ) );
    }

    return imgReport;
}

///////////////////////////////////////////////////////////////////////////////

void ImageReportFactory::writeHeader( std::ostream& os )
{
    // Write the collection header
    _header->write( os );

    // Write the HVC histogram indexer, if used
    if( _hvcHistogramIndexer ) {
        dynamic_pointer_cast< const KMeansPixelIndexer>(
            _hvcHistogramIndexer )->write( os );
    }
    
    // Write the correlogram indexer, if used and if its colorspace is HVC
    if( _corrDistances &&
        ( _correlogramIndexer->getTargetColorSpace() == ColorSpaces::HVC ) ) {
        dynamic_pointer_cast< const KMeansPixelIndexer>(
            _correlogramIndexer )->write( os );
    }
}
