// 
//  kMeansPixelIndexer.cpp
//  IFT630 - Processus concurrents et parallélisme
//  Benoit Gagnon (05 634 691)
//  
//  $Id: kMeansPixelIndexer.cpp 2926 2008-04-13 22:45:08Z benoit $
// 

#include "color.h"
#include "colorUtil.h"
#include "image.h"
#include "kMeansPixelIndexer.h"
#include "staticPixelIndexer.h"
#include "pixelDepths.h"

#include <algorithm>
#include <assert.h>
#include <iostream>
#include <math.h>
#include <time.h>

///////////////////////////////////////////////////////////////////////////////

struct BinCenterWriter
{
    BinCenterWriter( std::ostream& os ) : _os( os ){}
    void operator()( const Color< STFloat32 >& c ){ c.write( _os ); }
    std::ostream& _os;
};

///////////////////////////////////////////////////////////////////////////////

void splitParams(
    int64_t        params,
    unsigned char& splitR,
    unsigned char& splitG,
    unsigned char& splitB,
    unsigned int&  nbClass
)
{
    //  val: |r|g|b|n   |d|
    // byte: |0|1|2|3456|7|
    unsigned char* params_char;
    params_char = reinterpret_cast< unsigned char* >( &params );

    splitR = params_char[0];
    splitG = params_char[1];
    splitB = params_char[2];
    nbClass = *( reinterpret_cast< unsigned int* >( params_char + 3 ) );
}

///////////////////////////////////////////////////////////////////////////////

KMeansPixelIndexer* KMeansPixelIndexer::fromParams( int64_t params )
{
    unsigned char splitR, splitG, splitB;
    unsigned int nbClass;

    splitParams( params, splitR, splitG, splitB, nbClass );

    return new KMeansPixelIndexer( splitR, splitG, splitB, nbClass );
}

///////////////////////////////////////////////////////////////////////////////

KMeansPixelIndexer* KMeansPixelIndexer::fromParams(
    std::istream& is,
    int64_t params
)
{
    unsigned char splitR, splitG, splitB;
    unsigned int nbClass;

    splitParams( params, splitR, splitG, splitB, nbClass );

    return new KMeansPixelIndexer( is, splitR, splitG, splitB, nbClass );
}

///////////////////////////////////////////////////////////////////////////////

KMeansPixelIndexer::KMeansPixelIndexer(
    unsigned short splitR,
    unsigned short splitG,
    unsigned short splitB,
    unsigned short nbClass
) : PixelIndexer( nbClass ),
    _splitR( splitR ),
    _splitG( splitG ),
    _splitB( splitB ),
    _nbClass( nbClass ),
    _vecClasses(),
    _binCenters(),
    _staticIndexer()
{
    _staticIndexer.reset( new StaticPixelIndexer(_splitR, _splitG, _splitB ) );
    computeKMeans();

    // Compute cross distances
    fillDistancesMatrix();
}

///////////////////////////////////////////////////////////////////////////////

KMeansPixelIndexer::KMeansPixelIndexer(
    std::istream&  is,
    unsigned short splitR,
    unsigned short splitG,
    unsigned short splitB,
    unsigned short nbClass
) : PixelIndexer( nbClass ),
    _splitR( splitR ),
    _splitG( splitG ),
    _splitB( splitB ),
    _nbClass( nbClass ),
    _vecClasses( splitR * splitG * splitB, 0 ),
    _binCenters(),
    _staticIndexer()
{
    _staticIndexer.reset( new StaticPixelIndexer(_splitR, _splitG, _splitB) );

    // Read the bin LUT
    is.read( reinterpret_cast< char* >( &( _vecClasses[0] ) ),
             sizeof( int ) * _vecClasses.size() );

    // Read the bin centers from the stream
    _binCenters.reserve( nbClass );
    float binCenter[3];
    for( int i = 0; i < nbClass; ++i ) {
        is.read( reinterpret_cast< char* >( binCenter ), sizeof( float ) * 3 );
        _binCenters.push_back(
            Color< STFloat32 >( ColorSpaces::HVC, binCenter ) );
    }

    // Compute cross distances
    fillDistancesMatrix();
}

///////////////////////////////////////////////////////////////////////////////

KMeansPixelIndexer::~KMeansPixelIndexer()
{}

///////////////////////////////////////////////////////////////////////////////

PixelIndexer::BinIndexStorageType KMeansPixelIndexer::getIndex(
    const PixelStorageType* pixel
) const
{
    // Use the LUT to find the pixel index
    return _vecClasses[ _staticIndexer->getIndex( pixel ) ];
}

///////////////////////////////////////////////////////////////////////////////

ColorSpaces::ColorSpace KMeansPixelIndexer::getSourceColorSpace() const
{
    return ColorSpaces::RGB;
}

///////////////////////////////////////////////////////////////////////////////

ColorSpaces::ColorSpace KMeansPixelIndexer::getTargetColorSpace() const
{
    return ColorSpaces::HVC;
}

///////////////////////////////////////////////////////////////////////////////

const Color< STFloat32 >&
KMeansPixelIndexer::getBinCenter( unsigned int bin ) const
{
    assert( bin < getNbBins() );
    return _binCenters[bin];
}

///////////////////////////////////////////////////////////////////////////////

void KMeansPixelIndexer::write( std::ostream& os ) const
{
    // Write bin LUT
    os.write( reinterpret_cast< const char* >( &( _vecClasses[0] ) ),
              sizeof( int ) * _vecClasses.size() );

    // Write bin centers
    std::for_each( _binCenters.begin(), _binCenters.end(),
                   BinCenterWriter( os ) );
}

///////////////////////////////////////////////////////////////////////////////

void KMeansPixelIndexer::computeKMeans()
{    
    srand ( time( NULL ) );
    
    // Division statique de l'espace RGB
       
    scoped_ptr< Image< STFloat32 > > splitRGB(
        dynamic_cast< Image< STFloat32 >* >(
            ImageBase::create( _splitR * _splitG * _splitB, 1,
                ColorSpaces::RGB, STFloat32 ) ) );
    
    // On met les valeurs des centres dans une image RGB pour la conversion
    for ( int i = 0; i < _splitR * _splitG * _splitB; ++i ) {
        splitRGB->setValue(_staticIndexer->getBinCenter(i).getValue(
                           ColorSpaces::R), i, 0, ColorSpaces::R);
        splitRGB->setValue(_staticIndexer->getBinCenter(i).getValue(
                           ColorSpaces::G), i, 0, ColorSpaces::G);
        splitRGB->setValue(_staticIndexer->getBinCenter(i).getValue(
                           ColorSpaces::B), i, 0, ColorSpaces::B);
    }

    // Conversion des centres en HVC
    scoped_ptr< Image< STFloat32 > > splitHVC(
        dynamic_cast< Image< STFloat32 >* >( ColorUtil::convert(
            splitRGB.get(), ColorSpaces::HVC, STFloat32 ) ) );

    // On convertit splitHVC en un vecteur, pour faciliter le traitement
    std::vector < float > vecCentresHVC;
    vecCentresHVC.reserve(_splitR * _splitG * _splitB 
                          * splitHVC->getNbChannels());

    for ( int i = 0; i < _splitR * _splitG * _splitB; ++i ) {
        vecCentresHVC.push_back(splitHVC->getValue(i,0,ColorSpaces::H));
        vecCentresHVC.push_back(splitHVC->getValue(i,0,ColorSpaces::V));
        vecCentresHVC.push_back(splitHVC->getValue(i,0,ColorSpaces::C));
    }

    // On cherche les valeurs min et max des composantes H, V, C
    float minH( std::numeric_limits< float >::max() ),
          maxH( std::numeric_limits< float >::min() );
    float minV( std::numeric_limits< float >::max() ),
          maxV( std::numeric_limits< float >::min() );
    float minC( std::numeric_limits< float >::max() ),
          maxC( std::numeric_limits< float >::min() );

    for (size_t i = 0; i < vecCentresHVC.size(); i += 3 ) {
        if (vecCentresHVC[i] < minH)
            minH = vecCentresHVC[i];
        if (vecCentresHVC[i] > maxH)
            maxH = vecCentresHVC[i];
        
        if (vecCentresHVC[i+1] < minV)
            minV = vecCentresHVC[i+1];
        if (vecCentresHVC[i+1] > maxV)
            maxV = vecCentresHVC[i+1];
            
        if (vecCentresHVC[i+2] < minC)
            minC = vecCentresHVC[i+2];
        if (vecCentresHVC[i+2] > maxC)
            maxC = vecCentresHVC[i+2];
    }
         
    // Application du K-Means
            
    std::vector< float > vecMoyennes( 3 * _nbClass );
    std::vector< float > vecMoyennesOld( 3 * _nbClass );
    std::vector< int > vecClass( _splitR * _splitG * _splitB );
    std::vector< int > vecCompteurClass( _nbClass );
            
    // -- Etape 0 --
    // Assignation de valeurs aleatoires aux centres des "nouvelles" classes
    // Ces valeurs se trouvent entre les min et max trouves plus haut
    for ( size_t i = 0; i < vecMoyennes.size(); i += 3 ) {
        vecMoyennes[i] = ( rand() / float(RAND_MAX) ) * ( maxH - minH ) + minH;
        vecMoyennes[i+1] = ( rand() / float(RAND_MAX) ) * ( maxV - minV ) + minV;
        vecMoyennes[i+2] = ( rand() / float(RAND_MAX) ) * ( maxC - minC ) + minC;
        
        vecMoyennesOld[i] = 0.0f;
        vecMoyennesOld[i+1] = 0.0f;
        vecMoyennesOld[i+2] = 0.0f;
    }
    
    bool moyennesEgales = false;
    while ( !moyennesEgales ) {
        vecMoyennesOld = vecMoyennes;

        // -- Etape 1 --
        // Assignation des centres a la "meilleure" classe
        for ( size_t i = 0; i < vecCentresHVC.size(); i += 3 ) {
            float distAct;
            float distMin = std::numeric_limits< float >::max();
            int indiceBin = 0;
            
            for ( size_t j = 0; j < vecMoyennes.size(); j += 3 ) {
                distAct = powf( vecCentresHVC[i] - vecMoyennes[j], 2 );
                distAct += powf( vecCentresHVC[i+1] - vecMoyennes[j+1], 2 );
                distAct += powf( vecCentresHVC[i+2] - vecMoyennes[j+2], 2 );
                
                if ( distAct < distMin ) {
                    distMin = distAct;
                    indiceBin = j;
                }
            }
            vecClass[i/3] = indiceBin/3;
        }
        
        // -- Etape 2 --
        // Reinitialisation des moyennes
        for ( size_t i = 0; i < vecMoyennes.size(); ++i ) {
            vecMoyennes[i] = 0;
        }
        
        for ( size_t i = 0; i < vecCompteurClass.size(); ++i ) {
            vecCompteurClass[i] = 0;
        }
        
        // -- Etape 3 --
        // Raffinement des moyennes
        for ( size_t i = 0; i < vecCentresHVC.size(); i += 3 ) {
            int classAct = vecClass[i/3];
            vecCompteurClass[classAct] += 1;
            vecMoyennes[3*classAct] += vecCentresHVC[i];
            vecMoyennes[3*classAct+1] += vecCentresHVC[i+1];
            vecMoyennes[3*classAct+2] += vecCentresHVC[i+2];
        }
        
        for ( size_t i = 0; i < vecMoyennes.size(); i += 3 ) {
            // int classAct = vecClass[i/3];
            if (vecCompteurClass[i/3] == 0) {
                for ( size_t j = i; j < i+3; j++) {
                    vecMoyennes[j] = vecMoyennesOld[j];
                }
            }
            else {
                for ( size_t j = i; j < i + 3; ++j ) {
                    vecMoyennes[j] = vecMoyennes[j] 
                                     / vecCompteurClass[i/3];
                }
            }
        }
        
        // -- Étape 5 --
        // Ajustement du critere d'arret
        const float epsilon = 0.00000000001f;
        moyennesEgales = true;
        for ( size_t i = 0; i < vecMoyennes.size(); i += 3 ) {
            for ( size_t j = i; j < i+3; j++) {
                if (pow(vecMoyennes[j] - vecMoyennesOld[j], 2) > epsilon) {
                    moyennesEgales = false;
                    j = i+3;
                    i = vecMoyennes.size();
                }
            }
        }
    }

    _vecClasses = vecClass;

    _binCenters.reserve( _nbClass );
    for( unsigned int bin = 0; bin < getNbBins(); ++bin ) {
        depth_traits< STFloat32 >::storage_type center[3];

        center[0] = vecMoyennes[3*bin];
        center[1] = vecMoyennes[3*bin+1];
        center[2] = vecMoyennes[3*bin+2];

        _binCenters.push_back(
            Color< STFloat32 >( getTargetColorSpace(), center ) );
    }
    
}
