// 
//  uiDigestFrame.cpp
//  IFT630 - Processus concurrents et parallélisme
//  Benoit Gagnon (05 634 691)
//  
//  $Id: uiDigestFrame.cpp 2940 2008-04-15 03:48:29Z benoit $
// 

#include "uiDigestFrame.h"
#include "uiDigestProgress.h"
#include "uiColorTab.h"
#include "uiTextureTab.h"
#include "collectionReport.h"
#include "histogramFeature.h"
#include "coOccurenceMatrixFeature.h"
#include "correlogramFeature.h"
#include "osUtils.h"
#include "fsUtils.h"

///////////////////////////////////////////////////////////////////////////////

enum
{
    ID_BROWSE_STATICBOX,
    ID_BROWSE_BUTTON,
    ID_BROWSE_PATHTXT,
    ID_SAVE_BUTTON,
    ID_BOOKCTRL
};

///////////////////////////////////////////////////////////////////////////////

BEGIN_EVENT_TABLE( UiDigestFrame, wxFrame )
    EVT_BUTTON( ID_BROWSE_BUTTON, UiDigestFrame::OnBrowse )
    EVT_BUTTON( ID_SAVE_BUTTON, UiDigestFrame::OnSave )
END_EVENT_TABLE()

///////////////////////////////////////////////////////////////////////////////

UiDigestFrame::UiDigestFrame()
 : wxFrame(
        NULL,
        wxID_ANY,
        _T("Image import settings"),
        wxDefaultPosition,
        wxSize( 470, 600 ),
        wxCAPTION | wxMINIMIZE_BOX | wxCLOSE_BOX )
{
    const int spacing( 10 );

    // Top level panel
    wxPanel* topLevelPanel = new wxPanel(
                                    this,
                                    wxID_ANY,
                                    wxDefaultPosition,
                                    wxSize( 470, 600) );

    // Vertical layout
    wxBoxSizer* topSizer = new wxBoxSizer( wxVERTICAL );

    // Browse button
    wxStaticBox* inputStaticBox(
      new wxStaticBox(
        topLevelPanel,
        ID_BROWSE_STATICBOX,
        _T("Source images") ) );
    wxStaticBoxSizer* inputStaticBoxSizer(
      new wxStaticBoxSizer( inputStaticBox, wxHORIZONTAL) );
    inputStaticBoxSizer->Add(
        new wxButton( topLevelPanel, ID_BROWSE_BUTTON, _T("Browse...") )
    );

    // Dir path text
    _dirPathTXT = new wxStaticText(
                     topLevelPanel,
                     ID_BROWSE_PATHTXT,
                     _T("no directory selected"),
                     wxDefaultPosition,
                     wxDefaultSize,
                     wxST_NO_AUTORESIZE );
    inputStaticBoxSizer->Add( _dirPathTXT, 0, wxLEFT | wxGROW, 10 );

    topSizer->Add(
        inputStaticBoxSizer,
        0,
        wxGROW | wxTOP | wxLEFT | wxRIGHT, spacing
    );

    // Option tabs
    wxPanel* panel = new wxPanel( topLevelPanel );
    _book = new wxBookCtrl( panel, ID_BOOKCTRL,
                            wxPoint( 0, 0 ), wxSize( 460, 480 ) );
    _uiColorTab = new UiColorTab( _book );
    _book->AddPage( _uiColorTab, _T("Color analysis"), false );
    _uiTextureTab = new UiTextureTab( _book );
    _book->AddPage( _uiTextureTab, _T("Texture analysis"), false );
    topSizer->Add( panel, 0, wxALIGN_CENTER | wxALL, spacing );

    // Save button
    _saveButton = new wxButton(
                     topLevelPanel,
                     ID_SAVE_BUTTON,
                     _T("Digest") );
    _saveButton->SetDefault();
    _saveButton->Disable();
    topSizer->Add(
        _saveButton, 0, wxALIGN_RIGHT | wxALL, 2*spacing
    );

    topLevelPanel->SetSizer( topSizer );

    // Don't allow frame to get smaller than what the sizer tells it
    // Also sets the initial size as calculated by the sizers
    topSizer->SetSizeHints( this );
}

///////////////////////////////////////////////////////////////////////////////

void UiDigestFrame::OnBrowse( wxCommandEvent &event )
{
    wxString defaultPath = _T("/");
    wxDirDialog dialog(
        this,
        wxT( "Select a directory" ),
        defaultPath,
        wxDD_NEW_DIR_BUTTON
    );

    if ( dialog.ShowModal() == wxID_OK ) {
        _dirPath = dialog.GetPath();
        _dirPathTXT->SetLabel( _dirPath );
        _saveButton->Enable();
    }
}

///////////////////////////////////////////////////////////////////////////////

void UiDigestFrame::OnSave( wxCommandEvent &event )
{
    shared_ptr< const CollectionReport::Header > header( generateHeader() );
        
    FSUtils::FileList list;
    FSUtils::getDirListing( getInputDirPath(), list );
    
    UiDigestProgress* progress = new UiDigestProgress( list, header );
    progress->Show( true );
}

///////////////////////////////////////////////////////////////////////////////

wxString UiDigestFrame::getInputDirPath() const
{
    return _dirPath;
}

///////////////////////////////////////////////////////////////////////////////

UiDigestFrame::Values UiDigestFrame::getValues() const
{
    Values vals;

    vals._colorValues = _uiColorTab->getValues();
    vals._textureValues = _uiTextureTab->getValues();

    return vals;
}

///////////////////////////////////////////////////////////////////////////////

shared_ptr< CollectionReport::Header > UiDigestFrame::generateHeader() const
{
    shared_ptr< CollectionReport::Header >
    header( new CollectionReport::Header );
    UiDigestFrame::Values vals( getValues() );

    // RGB histo
    if( vals._colorValues._generateRGB ) {
        ImageFeature::Descriptor desc;

        desc._type = ImageFeature::HISTOGRAM;
        desc._colorSpace = ColorSpaces::RGB;
        desc._parameters = vals._colorValues._rgbValues.toParams();
        desc._valueSize = sizeof( HistogramFeature::StorageType );
        desc._nbValues = vals._colorValues._rgbValues.nbBins();

        header->add( desc );
    }

    // HVC histo
    if( vals._colorValues._generateHVC ) {
        ImageFeature::Descriptor desc;

        desc._type  = ImageFeature::HISTOGRAM;
        desc._colorSpace = ColorSpaces::HVC;
        desc._parameters = vals._colorValues._hvcValues.toParams();
        desc._valueSize  = sizeof( HistogramFeature::StorageType );
        desc._nbValues   = vals._colorValues._hvcValues._nbBins;

        header->add( desc );
    }

    // CoOccurence
    if( vals._textureValues._generateCoOcc ) {
        ImageFeature::Descriptor desc;

        desc._type  = ImageFeature::CO_OCCURENCE;
        desc._colorSpace = ColorSpaces::Grayscale;
        desc._parameters = vals._textureValues._distCoOcc;
        desc._valueSize = sizeof( float );
        desc._nbValues = CoOccurenceMatrixFeature::getNbStatistics();

        header->add( desc );
    }

    // Correlogram
    if( vals._textureValues._generateCorr ) {
        ImageFeature::Descriptor desc;

        desc._type  = ImageFeature::CORRELOGRAM;

        if( vals._textureValues._useRGB ) {
            desc._colorSpace = ColorSpaces::RGB;
            desc._parameters = vals._textureValues._rgbValues.toParams();
        }
        else {
            desc._colorSpace = ColorSpaces::HVC;
            desc._parameters = vals._textureValues._hvcValues.toParams();
        }
        // The 8th byte of the params is used for distances
        desc._parameters |= vals._textureValues._distCorr;
        desc._nbValues = CorrelogramFeature::getNbStatistics();
        desc._valueSize = sizeof( float );

        header->add( desc );
    }

    return header;
}
