// 
//  uiSearchFrame.cpp
//  IFT630 - Processus concurrents et parallélisme
//  Benoit Gagnon (05 634 691)
//  
//  $Id: uiSearchFrame.cpp 2940 2008-04-15 03:48:29Z benoit $
// 

#include <wx/wxprec.h>

#ifndef WX_PRECOMP
#include <wx/wx.h>
#endif

#include <wx/dir.h>
#include <wx/mstream.h>

#include "uiSearchFrame.h"
#include "uiGridItem.h"
#include "uiDigestFrame.h"
#include "wxThumbnailCtrl.h"
#include "wxThumbnailEvent.h"
#include "uiSearchOptions.h"
#include "collectionReport.h"
#include "imageReportFactory.h"
#include "randomRanker.h"
#include "histogramRanker.h"
#include "histogramFeature.h"
#include <vector>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <set>

///////////////////////////////////////////////////////////////////////////////

// Embedded resources
#include "new_png.h"
#include "open_png.h"
#include "random_png.h"
#include "search_png.h"

///////////////////////////////////////////////////////////////////////////////

#define MIN( A, B ) ( ( A ) < ( B ) ? ( A ) : ( B ) )

///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////

enum {
    ToolBarNew,
    ToolBarLoad,
    ToolBarRandom,
    ToolBarRetrieve
};

///////////////////////////////////////////////////////////////////////////////

IMPLEMENT_CLASS( UiSearchFrame, wxFrame )

///////////////////////////////////////////////////////////////////////////////

BEGIN_EVENT_TABLE( UiSearchFrame, wxFrame )
    EVT_IDLE( UiSearchFrame::OnIdle )
    EVT_MENU( ToolBarNew, UiSearchFrame::OnToolbarNew )
    EVT_MENU( ToolBarLoad, UiSearchFrame::OnToolbarLoad )
    EVT_MENU( ToolBarRandom, UiSearchFrame::OnRandom )
    EVT_MENU( ToolBarRetrieve, UiSearchFrame::OnToolbarRetrieveSimilar )
    EVT_SLIDER( ID_RESULTSSLIDER, UiSearchFrame::OnNbResultsChanged )
    EVT_SLIDER( ID_SLIDER, UiSearchFrame::OnZoom )

    EVT_THUMBNAIL_ITEM_SELECTED(wxID_ANY, UiSearchFrame::OnItemSelect)
    EVT_THUMBNAIL_ITEM_DESELECTED(wxID_ANY, UiSearchFrame::OnItemDeSelect)
    EVT_THUMBNAIL_LEFT_CLICK(wxID_ANY, UiSearchFrame::OnItemClick)
    EVT_THUMBNAIL_LEFT_DCLICK(wxID_ANY, UiSearchFrame::OnItemDoubleClick)
    EVT_THUMBNAIL_RIGHT_CLICK(wxID_ANY, UiSearchFrame::OnItemRightClick)
    EVT_CONTEXT_MENU(UiSearchFrame::OnItemContextClick)
END_EVENT_TABLE()

///////////////////////////////////////////////////////////////////////////////

inline wxBitmap loadPNG( const unsigned char* var, size_t sz )
{
    wxMemoryInputStream istream( var, sz );
    return wxBitmap( wxImage( istream ) );
}

/////////////////////////////////////////////////////////////////////////////

UiSearchFrame::UiSearchFrame(
    wxWindow*       parent,
    wxWindowID      id,
    const wxString& caption,
    const wxPoint&  pos,
    const wxSize&   size,
    long            style
) : _resultCountSlider( 0 ),
    _scrolledWindow( 0 ),
    _zoomSlider( 0 ),
    _searchOptions( 0 ),
    _scaleFactor( 1.0f ),
    _selectedItem( 0 ),
    _collection()
{
    wxFrame::Create( parent, id, caption, pos, size, style );

    CreateControls();
    GetSizer()->Fit( this );
    GetSizer()->SetSizeHints( this );

    _scaleFactor = _zoomSlider->GetValue() / 100.0f;
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::SetCollection(
    const shared_ptr< const CollectionReport::Header >& header,
    const shared_ptr< CollectionReport >&               collection
)
{
    _collection = collection;
    _searchOptions->SetCollectionHeader( header );
    _resultCountSlider->SetRange(
        1,
        MIN( 100, _collection->getNbImages() )
    );
    _resultCountSlider->SetValue(
        MIN( _collection->getNbImages(), 20 ) );

    // Enable toolbar buttons
    GetToolBar()->EnableTool( ToolBarRandom, true );

	// Enable nb of results slider
	_resultCountSlider->Enable();

    // Display random results
    _currentResults.clear();
    _collection->collect( _currentResults, RandomRanker() );
    displayGridResults();
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::CreateControls()
{
    wxToolBar* itemToolBar12 =
        CreateToolBar( wxTB_FLAT | wxTB_TEXT | wxTB_HORIZONTAL, ID_TOOLBAR1 );
    itemToolBar12->SetToolBitmapSize( wxSize( 32, 32 ) );

    // New collection tool
    itemToolBar12->AddTool(
        ToolBarNew,
        _("New"),
        loadPNG( new_png, sizeof( new_png ) ),
        wxNullBitmap,
        wxITEM_NORMAL,
        _("new image collection"),
        _("new image collection")
    );

    // Load collection tool
    itemToolBar12->AddTool(
        ToolBarLoad,
        _("Open"),
        loadPNG( open_png, sizeof( open_png ) ),
        wxNullBitmap,
        wxITEM_NORMAL,
        _("open an image collection"),
        _("open an image collection")
    );
    itemToolBar12->AddSeparator();

    // Get random images tool
    itemToolBar12->AddTool(
        ToolBarRandom,
        _("Randomize"),
        loadPNG( random_png, sizeof( random_png ) ),
        wxNullBitmap,
        wxITEM_NORMAL,
        _("get random images"),
        _("get random images")
    );
    itemToolBar12->EnableTool( ToolBarRandom, false );

    // Retrieve similar tool
    itemToolBar12->AddTool(
        ToolBarRetrieve,
        _("Search"),
        loadPNG( search_png, sizeof( search_png ) ),
		wxNullBitmap,
        wxITEM_NORMAL,
		_("search for similar images"),
        _("search for similar images")
    );

    itemToolBar12->AddSeparator();
    itemToolBar12->Realize();

    // Main sizer
    wxBoxSizer* itemBoxSizer23 = new wxBoxSizer( wxVERTICAL );

    // Thumbnail browser
    _scrolledWindow = new wxThumbnailCtrl(
        this,
        ID_SCROLLEDWINDOW1,
        wxDefaultPosition,
        wxSize( 800, 400 ),
        wxSUNKEN_BORDER | wxHSCROLL | wxVSCROLL
    );
    _scrolledWindow->SetScrollbars( 1, 1, 1000, 1000 );
    _scrolledWindow->SetBackgroundColour( wxColour( 40, 40, 40 ) );
    _scrolledWindow->SetFocusRectColour( wxColour( 255, 255, 255 ) );
    _scrolledWindow->SetFont( *wxSMALL_FONT );
    itemBoxSizer23->Add( _scrolledWindow, 1, wxGROW | wxALL, 5 );

    // Horizontal sizer
    wxBoxSizer* itemBoxSizer25 = new wxBoxSizer(wxHORIZONTAL);

	// Number of results (label)
    wxStaticText* itemStaticText21(
        new wxStaticText(
            this,
            wxID_STATIC,
            _("number of results: ")
        )
    );
    itemBoxSizer25->Add( itemStaticText21, 0, wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL, 5 );

    // Number of results (slider)
	itemToolBar12->Realize();
    _resultCountSlider = new wxSlider(
        this,
        ID_RESULTSSLIDER,
        10,
        1,
        10,
        wxDefaultPosition,
        wxSize( 200, 32 ),
        wxSL_HORIZONTAL | wxNO_BORDER
    );
	_resultCountSlider->Disable();
	itemBoxSizer25->Add( _resultCountSlider, 0, wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL, 5 );

    // Zoom label
    wxStaticText* itemStaticText26 =
        new wxStaticText(
            this,
            wxID_STATIC,
			_("thumbnail size:") );
    itemBoxSizer25->Add( itemStaticText26, 0, wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL, 5);

    // Zoom slider
    _zoomSlider =
        new wxSlider(
            this,
            ID_SLIDER,
            100,
            20,
            100,
            wxDefaultPosition,
            wxSize( 100, 32 ),
            wxNO_BORDER );
    itemBoxSizer25->Add( _zoomSlider, 0, wxALIGN_RIGHT | wxALIGN_CENTER_VERTICAL, 5 );

	// Add horizontal slider
	itemBoxSizer23->Add( itemBoxSizer25, 0, wxALIGN_RIGHT|wxALL, 5 );

    // Search options
    _searchOptions = new UiSearchOptions( this );
    _searchOptions->Disable();
    itemBoxSizer23->Add( _searchOptions, 0, wxALL, 10 );

    // Set sizer
	itemBoxSizer23->SetSizeHints( this );
    this->SetSizer( itemBoxSizer23 );
	this->SetToolBar( itemToolBar12 );
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::displayGridResults()
{
    _scrolledWindow->Freeze();
    _scrolledWindow->Clear();

    int maxX = 0;
    int maxY = 0;
    int i = 0;
    int maxCount = _resultCountSlider->GetValue();
    for ( std::multiset< SearchResult >::const_iterator
          it( _currentResults.begin() );
          ( i < maxCount ) && ( it != _currentResults.end() ); ++it  ) {
        UiGridItem* tmpItem = new UiGridItem( *it , _scaleFactor);
        wxSize tmpSize = tmpItem->DoGetBestSize();
        if (maxX < tmpSize.x) maxX = tmpSize.x;
        if (maxY < tmpSize.y) maxY = tmpSize.y;
        _scrolledWindow->Append( tmpItem );
        ++i;
    }

    _scrolledWindow->SetThumbnailImageSize( wxSize( maxX, maxY ) );

    _scrolledWindow->Thaw();
}

///////////////////////////////////////////////////////////////////////////////

// Should we show tooltips?
bool UiSearchFrame::ShowToolTips()
{
    return true;
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnToolbarNew( wxCommandEvent& event )
{
    UiDigestFrame* digestFrame = new UiDigestFrame();
    digestFrame->Show( true );

    // Close this window if there's no collection
    if( !_collection ) {
        Close( true );
    }
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnToolbarLoad( wxCommandEvent& event )
{
    wxString collectionPath;
    wxString defaultPath = _T("/");
    wxString wildcard = _T( "COL files (*.col)|*.col" );

    wxFileDialog dialog(
        this,
        _T( "Open collection" ),
        defaultPath,
        _T( "" ),
        wildcard,
        wxOPEN | wxFILE_MUST_EXIST
    );

    wxString reportSavePath;
    if ( dialog.ShowModal() == wxID_OK ) {
        collectionPath = dialog.GetPath();
    }
    else {
        return;
    }

    // Open the collection file
    std::ifstream file;
    file.open(
        collectionPath.fn_str(),
        std::ifstream::in | std::ifstream::binary
    );
    assert( file.good() );

    // Generate header
    shared_ptr< CollectionReport::Header > header(
        new CollectionReport::Header( file ) );
    ImageReportFactory factory( header, file );

    // Empty collection
    shared_ptr< CollectionReport > collection( new CollectionReport() );
    while( file.good() && file.peek() != EOF )
    {
        collection->insert( factory.create( file ) );
    }

    // Close file
    file.close();

    // Set the collection
    SetCollection( header, collection );
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnRandom( wxCommandEvent& event )
{
    _currentResults.clear();
    _collection->collect( _currentResults, RandomRanker() );
    displayGridResults();
    GetToolBar()->EnableTool( ToolBarRetrieve, false );
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnMenuFileExitClick( wxCommandEvent & event )
{
    Close( true );
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnZoom( wxCommandEvent& event )
{
    _scaleFactor = (float) _zoomSlider->GetValue() / 100.f;

    _scrolledWindow->Freeze();
        
    int maxX = 0;
    int maxY = 0;
    for(int i=0; i < _scrolledWindow->GetCount(); ++i) {
        UiGridItem* tmpItem = (UiGridItem*) _scrolledWindow->GetItem( i );
        tmpItem->setScale( _scaleFactor );
        wxSize tmpSize = tmpItem->DoGetBestSize();
        if ( maxX < tmpSize.x )
            maxX = tmpSize.x;
        if ( maxY < tmpSize.y )
            maxY = tmpSize.y;
    }

    _scrolledWindow->SetThumbnailImageSize( wxSize( maxX, maxY ) );
    _scrolledWindow->Thaw();
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnItemSelect( wxThumbnailEvent& event ) {
    _selectedItem = (UiGridItem*) _scrolledWindow->GetItem(event.GetIndex());
    GetToolBar()->EnableTool( ToolBarRetrieve, true );
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnItemDeSelect( wxThumbnailEvent& event )
{
    _selectedItem = 0;
    GetToolBar()->EnableTool( ToolBarRetrieve, false );
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnItemClick( wxThumbnailEvent& event )
{
    // int itemID = event.GetIndex();
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnItemDoubleClick( wxThumbnailEvent& event )
{
    int itemID = event.GetIndex();

    // Get the selected item
    UiGridItem* selectedItem( dynamic_cast< UiGridItem* >(
        _scrolledWindow->GetItem( itemID ) ) );

    // Retrieve similar
    retrieveSimilar( selectedItem );
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnItemRightClick( wxThumbnailEvent& event )
{
    _selectedItem = (UiGridItem*) _scrolledWindow->GetItem( event.GetIndex() );
    GetToolBar()->EnableTool( ToolBarRetrieve, true );

    ContextMenu();
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnItemContextClick( wxContextMenuEvent& event )
{
    if ( _selectedItem != 0 ) {
      //if the context menu doesn't appear close to the selected image
      // -> calculate position and give it as parameter
      ContextMenu();
    }
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::ContextMenu(
    const wxPoint& pos /* = wxDefaultPosition */
)
{
    wxMenu* contextMenu = new wxMenu;
    contextMenu->Append(
        ToolBarRetrieve, _("Retrieve similar..."), _T(""), wxITEM_NORMAL);
    wxWindow* popupContextMenu = new wxWindow(this, wxID_ANY);
    popupContextMenu->PopupMenu(contextMenu, pos);
    popupContextMenu->Show(false);
    delete popupContextMenu;
    delete contextMenu;
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnNbResultsChanged( wxCommandEvent& event )
{
    _scrolledWindow->Freeze();
    int prevCount = _scrolledWindow->GetCount();
    int newCount = _resultCountSlider->GetValue();
    if( newCount > prevCount ) {
        nbResultsUp();
    }
    else if ( newCount < prevCount ) {
        nbResultsDown();
    }
    wxCommandEvent dummy;
    OnZoom( dummy );
    _scrolledWindow->Thaw();
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::nbResultsUp()
{
    //FIXME This is buggy when the slider moves to fast

    // Advance to the current position
    std::multiset< SearchResult >::const_iterator it =
        _currentResults.begin();
    unsigned int pos( _resultCountSlider->GetValue() - 1 );
    for( size_t i = 0; i < pos; ++i ) {
        ++it;
    }

    // Add new grid items
    int currentCount = _scrolledWindow->GetCount();
    int newCount = _resultCountSlider->GetValue();
    for( int i=0;
        it != _currentResults.end() && i < (newCount - currentCount);
        ++i) {
        UiGridItem* tmpItem = new UiGridItem( *it, _scaleFactor);
        _scrolledWindow->Append( tmpItem );
        ++it;
    }
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::nbResultsDown()
{
    //FIXME This is buggy when the slider moves to fast

    int currentCount = _scrolledWindow->GetCount();
    int newCount = _resultCountSlider->GetValue();

    // Remove grid items
    for( int pos = currentCount; pos >= newCount; --pos ) {
        _scrolledWindow->Delete( pos );
    }
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnToolbarRetrieveSimilar( wxCommandEvent& event )
{
    if ( _selectedItem != 0 ) {
        // Retrieve similar
        retrieveSimilar( _selectedItem );
    }
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::OnIdle( wxIdleEvent& event )
{
    if( _selectedItem == 0 ){ 
        GetToolBar()->EnableTool( ToolBarRetrieve, false );
    }
}

///////////////////////////////////////////////////////////////////////////////

void UiSearchFrame::retrieveSimilar( UiGridItem* item )
{
    // Clear the current results
    _currentResults.clear();

    // Generate an ImageRanker based on current search options
    shared_ptr< ImageRanker > ranker(
        _searchOptions->generateImageRanker( item->getImageReport() ) );

    // Start a search
    _collection->collect( _currentResults, *ranker );

    // Display the results
    displayGridResults();

    // Disable the "retrieve similar" tool
    GetToolBar()->EnableTool( ToolBarRetrieve, false );
}
